import uuid
from .base58 import Base58
from .uuid7 import UUID, new_uuid7

_base58 = Base58()


def uuid1(node=None, clock_seq=None):
    # type: (int, int) -> UUID58
    return UUID58(uuid.uuid1(node, clock_seq).hex)


def uuid3(namespace, name):
    # type: (uuid.UUID, str) -> UUID58
    return UUID58(uuid.uuid3(namespace, name).hex)


def uuid4():
    # type: () -> UUID58
    return UUID58(uuid.uuid4().hex)


def uuid5(namespace, name):
    # type: (uuid.UUID, str) -> UUID58
    return UUID58(uuid.uuid5(namespace, name).hex)


def uuid7():
    # type: () -> UUID58
    return UUID58(new_uuid7().hex)


class UUID58(UUID):
    @property
    def base58(self):
        # type: () -> str
        return _base58.encode(self.bytes).decode('ascii')

    @property
    def base58check(self):
        # type: () -> str
        return _base58.encode_check(self.bytes).decode('ascii')

    @classmethod
    def from_base58(cls, s):
        # type: (str) -> 'UUID58'
        return UUID58(bytes=_base58.decode(s.encode('ascii')))

    @classmethod
    def from_base58check(cls, s):
        # type: (str) -> 'UUID58'
        return UUID58(bytes=_base58.decode_check(s.encode('ascii')))



import sys
from hashlib import sha256

BITCOIN_ALPHABET = b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
RIPPLE_ALPHABET = b'rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz'
FLICKR_ALPHABET = b'123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ'

_py3k = sys.version_info.major > 2
_b58size = 58


class Base58:
    _b58tab = _b58rev = None

    def __init__(self, ripple=False, flickr=False):
        # type: (bool, bool) -> None
        if ripple:
            self._b58alphabet = RIPPLE_ALPHABET
        elif flickr:
            self._b58alphabet = FLICKR_ALPHABET
        else:
            self._b58alphabet = BITCOIN_ALPHABET

    def encode(self, s):
        # type: (bytes) -> bytes
        if self._b58tab is None:
            self._b58tab = [bytes((i,)) for i in self._b58alphabet] \
                if _py3k else [bytes(i) for i in self._b58alphabet]
        pad = len(s)
        s = s.lstrip(b'\0')
        pad -= len(s)
        p, acc = 1, 0
        if _py3k:
            for c in reversed(s):
                acc += p * c
                p <<= 8
        else:
            for c in reversed(s):
                acc += p * ord(c)
                p <<= 8
        buf = b''
        b58tab = self._b58tab
        while acc:
            acc, i = divmod(acc, _b58size)
            buf = b58tab[i] + buf
        return b58tab[0] * pad + buf

    def encode_check(self, s):
        # type: (bytes) -> bytes
        s += sha256(sha256(s).digest()).digest()[:4]
        return self.encode(s)

    def decode(self, s):
        # type: (bytes) -> bytes
        if self._b58tab is None:
            self._b58tab = [bytes((i,)) for i in self._b58alphabet] \
                if _py3k else [bytes(i) for i in self._b58alphabet]
        if self._b58rev is None:
            self._b58rev = {v: k for k, v in enumerate(self._b58alphabet)}
        pad = len(s)
        s = s.lstrip(self._b58tab[0])
        pad -= len(s)
        b58rev = self._b58rev
        acc = 0
        buf = b''
        for c in s:
            acc = acc * _b58size + b58rev[c]
        if _py3k:
            buf_ = []
            while acc:
                acc, c = divmod(acc, 256)
                buf_.append(c)
            buf = bytes(reversed(buf_))
        else:
            while acc:
                acc, c = divmod(acc, 256)
                buf = chr(c) + buf
        return b'\0' * pad + buf

    def decode_check(self, s):
        # type: (bytes) -> bytes
        buf = self.decode(s)
        buf, check = buf[:-4], buf[-4:]
        if check != sha256(sha256(buf).digest()).digest()[:4]:
            raise ValueError('Invalid checksum')
        return buf



import sys
import binascii

BASE32_ALPHABET = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
BASE32HEX_ALPHABET = b'0123456789ABCDEFGHIJKLMNOPQRSTUV'
CROCKFORD_ALPHABET = b'0123456789ABCDEFGHJKMNPQRSTVWXYZ'
ZBASE32_ALPHABET = b'ybndrfg8ejkmcpqxotluwisza345h769'

_py3k = sys.version_info.major > 2


class Base32:
    _b32tab = _b32rev = None

    def __init__(self, base32hex=False, crockford=False, zbase32=False):
        # type: (bool, bool, bool) -> None
        if base32hex:
            self._b32alphabet = BASE32HEX_ALPHABET
            self._padding = True
        elif crockford:
            self._b32alphabet = CROCKFORD_ALPHABET
            self._padding = False
        elif zbase32:
            self._b32alphabet = ZBASE32_ALPHABET
            self._padding = False
        else:
            self._b32alphabet = BASE32_ALPHABET
            self._padding = True

    def encode(self, s):
        # type: (bytes) -> bytes
        if self._b32tab is None:
            b32tab = [bytes((i,)) for i in self._b32alphabet] \
                if _py3k else [bytes(i) for i in self._b32alphabet]
            self._b32tab = [a + b for a in b32tab for b in b32tab]
        leftover = len(s) % 5
        if leftover:
            s = s + b'\0' * (5 - leftover)
        b32tab = self._b32tab
        encoded = bytearray()
        if _py3k:
            from_bytes = int.from_bytes
            for i in range(0, len(s), 5):
                c = from_bytes(s[i:i+5], 'big')
                encoded += (b32tab[c >> 30] +
                            b32tab[(c >> 20) & 0x3ff] +
                            b32tab[(c >> 10) & 0x3ff] +
                            b32tab[c & 0x3ff])
        else:
            for i in range(0, len(s), 5):
                c = int(binascii.hexlify(s[i:i+5]), 16)
                encoded += (b32tab[c >> 30] +
                            b32tab[(c >> 20) & 0x3ff] +
                            b32tab[(c >> 10) & 0x3ff] +
                            b32tab[c & 0x3ff])
        if leftover == 1:
            encoded[-6:] = b'======' if self._padding else b''
        elif leftover == 2:
            encoded[-4:] = b'====' if self._padding else b''
        elif leftover == 3:
            encoded[-3:] = b'===' if self._padding else b''
        elif leftover == 4:
            encoded[-1:] = b'=' if self._padding else b''
        return bytes(encoded)

    def decode(self, s):
        # type: (bytes) -> bytes
        if self._b32rev is None:
            self._b32rev = {v: k for k, v in enumerate(self._b32alphabet)}
        leftover = len(s) % 8
        if leftover:
            s = s + b'=' * (8 - leftover)
        size = len(s)
        s = s.rstrip(b'=')
        pad = size - len(s)
        acc = 0
        b32rev = self._b32rev
        decoded = bytearray()
        if _py3k:
            for i in range(0, len(s), 8):
                quanta = s[i: i + 8]
                acc = 0
                try:
                    for c in quanta:
                        acc = (acc << 5) + b32rev[c]
                except KeyError:
                    raise binascii.Error('Non-base32 digit found')
                decoded += acc.to_bytes(5, 'big')
            if size % 8 or pad not in {0, 1, 3, 4, 6}:
                raise binascii.Error('Incorrect padding')
            if pad and decoded:
                acc <<= 5 * pad
                last = acc.to_bytes(5, 'big')
                leftover = (43 - 5 * pad) // 8
                decoded[-5:] = last[:leftover]
        else:
            for i in range(0, len(s), 8):
                quanta = s[i: i + 8]
                acc = 0
                try:
                    for c in quanta:
                        acc = (acc << 5) + b32rev[c]
                except KeyError:
                    raise binascii.Error('Non-base32 digit found')
                decoded += binascii.unhexlify('%010x' % acc)
            if size % 8 or pad not in {0, 1, 3, 4, 6}:
                raise binascii.Error('Incorrect padding')
            if pad and decoded:
                acc <<= 5 * pad
                last = binascii.unhexlify('%010x' % acc)
                leftover = (43 - 5 * pad) // 8
                decoded[-5:] = last[:leftover]
        return bytes(decoded)



import time
import uuid

_last_monotonicity_timestamp = 0
_time_ns = lambda: int(time.time() * 1000000000)
if hasattr(time, 'time_ns'):
    _time_ns = time.time_ns


class UUID(uuid.UUID):
    @property
    def time(self):
        # type: () -> int
        if self.version == 7:
            return int((self.int >> 80) / 1000)
        return super(UUID, self).time

    @property
    def time_ms(self):
        # type: () -> int
        return int(self.int >> 80)

    @property
    def time_ns(self):
        # type: () -> int
        return int(((self.int >> 80) * 1000000) + int((((self.int >> 64) & 0xfff) * 1000000) / 4096))

    @property
    def counter_seq(self):
        # type: () -> int
        return int((self.int >> 64) & 0xfff)


def _get_monotonicity_timestamp():
    # type: () -> (int, int)
    global _last_monotonicity_timestamp

    ns = _time_ns()
    ms, seq = divmod(ns, 1000000)
    ms &= 0xffffffffffff
    # RFC9562: Replace Leftmost Random Bits with Increased Clock Precision (Method 3)
    seq = int((float(seq) / 1000000) * 4096)
    ts = (ms << 12) + seq
    if ts <= _last_monotonicity_timestamp:
        ts = _last_monotonicity_timestamp + 1
        ms = (ts >> 12) & 0xffffffffffff
        seq = ts & 0xfff

    _last_monotonicity_timestamp = ts
    return ms, seq


def new_uuid7():
    # type: () -> UUID
    ms, seq = _get_monotonicity_timestamp()
    seq |= 0x7000
    return UUID(int=(((ms << 16) + seq) << 64) + (uuid.uuid4().int & 0xffffffffffffffff))



import os
import sys
import time
import struct
import hashlib
import platform
from threading import Lock
from datetime import datetime
from .base32 import Base32


class InvalidId(Exception):
    pass


def _real_machine_id():
    # type: () -> bytes
    hid = ''
    system = platform.system()
    if system == 'Linux':
        try:
            with open('/sys/class/dmi/id/product_uuid') as f:
                hid = f.read()
        except IOError:
            pass
    # TODO: readPlatformMachineID is Linux only implementation. Implement other OS if needed
    if not hid:
        hid = platform.node()
    if hid:
        md5 = hashlib.md5()
        md5.update(hid.encode('latin-1'))
        return md5.digest()[:3]
    # Fallback to rand number if machine id can't be gathered
    return os.urandom(3)


def _count_generator():
    def rand_int():
        # type: () -> int
        buf = struct.unpack('BBB', os.urandom(3))
        return buf[0] << 16 | buf[1] << 8 | buf[2]
    count = rand_int()
    while True:
        # reset 3-byte counter
        if count < 0xffffff:
            count += 1
        else:
            count = 0
        yield count


_py3k = sys.version_info.major > 2
_encoded_len = 20
_raw_len = 12
_lock = Lock()
_base32hex = Base32(base32hex=True)
_machine_id = _real_machine_id()
_object_id_counter = _count_generator()


def new_xid():
    # type: () -> Xid
    now = int(time.time())
    pid = os.getpid()
    id_ = b''
    # Timestamp, 4 bytes, big endian
    id_ += struct.pack('>I', now & 0xffffffff)
    # Machine, first 3 bytes of md5(hostname)
    id_ += _machine_id
    # Pid, 2 bytes, specs don't specify endianness, but we use big endian.
    id_ += struct.pack('>H', pid & 0xffff)
    # Increment, 3 bytes, big endian
    _lock.acquire()
    count = next(_object_id_counter)
    _lock.release()
    id_ += struct.pack('>I', count & 0xffffff)[-3:]
    return Xid(id_)


class Xid:
    def __init__(self, id_):
        # type: (bytes) -> None
        if len(id_) != _raw_len:
            raise InvalidId('Xid length is {} bytes'.format(_raw_len))
        self._value = id_

    @property
    def time(self):
        # type: () -> datetime
        time_, = struct.unpack('>I', self._value[:4])
        return datetime.fromtimestamp(time_)

    @property
    def machine(self):
        # type: () -> bytes
        return self._value[4:7]

    @property
    def pid(self):
        # type: () -> int
        pid, = struct.unpack('>H', self._value[7:9])
        return pid

    @property
    def counter(self):
        # type: () -> int
        count, = struct.unpack('>I', b'\x00' + self._value[9:12])
        return count

    @property
    def string(self):
        # type: () -> str
        str_ = _base32hex.encode(self._value).lower()[:_encoded_len]
        return str_.decode() if _py3k else str_

    @property
    def bytes(self):
        # type: () -> bytes
        return self._value

    def __repr__(self):
        return '<Xid at {}>'.format(self.string)

    def __str__(self):
        return self.string

    def __eq__(self, other):
        # type: ('Xid') -> bool
        return self._value == other._value

    def __ne__(self, other):
        # type: ('Xid') -> bool
        return self._value != other._value

    def __lt__(self, other):
        # type: ('Xid') -> bool
        return self._value < other._value

    def __le__(self, other):
        # type: ('Xid') -> bool
        return self._value <= other._value

    def __gt__(self, arg):
        # type: ('Xid') -> bool
        return self._value > arg._value

    def __ge__(self, other):
        # type: ('Xid') -> bool
        return self._value >= other._value

    @classmethod
    def from_string(cls, s):
        # type: (str) -> 'Xid'
        if len(s) != _encoded_len:
            raise InvalidId('Xid string is {} characters'.format(_encoded_len))
        val = _base32hex.decode(s.upper().encode('ascii'))
        return cls(val)



from setuptools import setup

setup(
    name='hisak-idtools',
    version='0.2.0',
    packages=['hisak.idtools'],
    package_dir={'hisak.idtools': 'src'},
    url='https://gitlab.com/HiSakDev/idtools',
    license='MIT',
    author='HiSakDev',
    author_email='sak.devac@gmail.com',
    description='Identifier generator tools often used privately',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)
